//! Deals with getting all paragraphs for a given law text

mod ris_structure;

use log::info;
use serde::Deserialize;

use crate::misc::{current_date, Error};

use ris_structure::OgdSearchResult;

/// Parses a law text from the Austrian RIS (Rechtsinformationssystem) based on the given `law_id`.
///
/// This function iterates over all pages of the law text, with each page containing a maximum of 100
/// paragraphs in XML format. It extracts and returns the links to each paragraph as XML URLs.
///
/// The first section of the first page is skipped (`skip` is set to true) because it always
/// contains the table of contents. The function continues processing subsequent pages until there
/// are no more pages left to fetch.
///
/// # Parameters
///
/// - `law_id`: The unique identifier of the law in the RIS system.
///
/// # Returns
///
/// - `Ok(Vec<String>)`: A vector of XML file links representing paragraphs from the given law text.
/// - `Err(Error)`: An error if there was an issue fetching or parsing the law text.
///
/// # Example
/// ```
/// use risp::risparser::overview::parse;
///
/// let list_with_xml_links_to_paragraphs = parse(10001905).unwrap();
/// assert_eq!(list_with_xml_links_to_paragraphs.len(), 31); // TEG has 31 paragraphs
/// assert_eq!(list_with_xml_links_to_paragraphs[0], "https://www.ris.bka.gv.at/Dokumente/Bundesnormen/NOR12025190/NOR12025190.xml"); // Link to first paragraph
/// ```
pub fn parse(law_id: usize) -> Result<Vec<String>, Error> {
    let mut page = 1;
    let mut skip = true;
    let mut ret = Vec::new();
    loop {
        info!("=== Fetching overview page #{page} ===");
        let json = fetch_page(law_id, page)?;
        let (cont, nodes) = parse_from_str(&json, skip)?;
        for n in nodes {
            ret.push(n.clone());
        }
        if !cont {
            break;
        }
        skip = false;
        page += 1;
    }

    Ok(ret)
}

#[derive(Deserialize)]
#[serde(rename_all = "PascalCase")]
struct Overview {
    ogd_search_result: OgdSearchResult,
}

fn parse_from_str(content: &str, skip_first: bool) -> Result<(bool, Vec<String>), Error> {
    let mut ret = Vec::new();
    let wrapper: Overview = serde_json::from_str(content)?;

    let iter = wrapper.ogd_search_result.get_par().into_iter();
    let boxed_iter: Box<dyn Iterator<Item = String>> = if skip_first {
        Box::new(iter.skip(1))
    } else {
        Box::new(iter)
    };
    for par in boxed_iter {
        ret.push(par);
    }

    if !wrapper.ogd_search_result.has_next_page() {
        return Ok((false, ret));
    }
    Ok((true, ret))
}

#[cfg(not(test))]
/// Fetches the json content of the given overview (`law_id`) from the RIS API.
///
/// # Errors
/// Fails if `ureq` can't create a connection, probably because there's no internet connection? (Or RIS is not online.)
fn fetch_page(overview_id: usize, page: usize) -> Result<String, Error> {
    Ok(
        ureq::post("https://data.bka.gv.at/ris/api/v2.6/Bundesrecht")
            .send_form(&[
                ("Applikation", "BrKons"),
                ("Gesetzesnummer", &format!("{overview_id}")),
                ("DokumenteProSeite", "OneHundred"),
                ("Seitennummer", &format!("{page}")),
                ("Fassung.FassungVom", &current_date()),
            ])?
            .into_string()?,
    )
}

#[cfg(test)]
fn fetch_page(overview_id: usize, page: usize) -> Result<String, Error> {
    use std::fs;

    let expected_filename = format!("./data/cache/law-{overview_id}-{page}");

    match fs::read_to_string(&expected_filename) {
        Ok(data) => Ok(data),
        Err(_) => {
            info!("Not finding law_id {overview_id} (page {page}) in the cache, downloading...");
            let data = ureq::post("https://data.bka.gv.at/ris/api/v2.6/Bundesrecht")
                .send_form(&[
                    ("Applikation", "BrKons"),
                    ("Gesetzesnummer", &format!("{overview_id}")),
                    ("DokumenteProSeite", "OneHundred"),
                    ("Seitennummer", &format!("{page}")),
                    ("Fassung.FassungVom", &current_date()),
                ])?
                .into_string()?;
            fs::write(expected_filename, &data).expect("Unable to write file");
            Ok(data)
        }
    }
}

#[cfg(test)]
mod tests {
    use crate::risparser::overview::parse;
    use pretty_assertions::assert_eq;
    use std::fs;

    #[test]
    fn overview() {
        let paths = fs::read_dir("./data/expected/overview/")
            .expect("No folder with expected overview files");

        for path in paths {
            let path = path.unwrap();
            let file_path = format!("{}", path.path().display());
            let law_id = path.file_name().into_string().unwrap().to_string();
            let law_id: usize = law_id
                .parse()
                .expect("Filename needs to be usize (= law_id in RIS)");

            let expected = fs::read_to_string(file_path).expect("Could not read file {file_path}.");
            let expected = expected.trim().split('\n').collect::<Vec<&str>>();

            let actual = parse(law_id).unwrap();
            assert_eq!(actual, expected);
        }
    }
}
