//! Deals with getting all paragraphs for a given law text
mod parser;

use log::info;

use crate::{law::LawBuilder, misc::Error};

use self::parser::Risdok;

pub struct Parser {
    remove: Vec<String>,
    replace: Vec<(String, String)>,
}

impl Parser {
    pub fn new() -> Self {
        Self {
            remove: Vec::new(),
            replace: Vec::new(),
        }
    }

    pub fn add_string_to_remove(&mut self, data: &str) {
        self.remove.push(data.into());
    }

    pub fn add_string_to_replace(&mut self, search: &str, replace: &str) {
        self.replace.push((search.into(), replace.into()));
    }

    /// Parses the content available in `url`. Calls appropriate functions in supplied `LawBuilder`.
    pub fn parse(&self, url: &str, builder: &mut LawBuilder) -> Result<bool, Error> {
        info!("Parsing {url}");
        let xml = fetch(url)?;

        self.parse_from_str(&xml, builder)
    }

    fn parse_from_str(&self, xml: &str, builder: &mut LawBuilder) -> Result<bool, Error> {
        let mut xml = String::from(xml);
        for r in &self.remove {
            xml = xml.replace(r, "");
        }
        for (search, replace) in &self.replace {
            xml = xml.replace(search, replace);
        }

        Risdok::from_str(&xml, builder)
    }
}

#[cfg(not(test))]
fn fetch(url: &str) -> Result<String, Error> {
    Ok(ureq::get(url).call()?.into_string()?)
}

#[cfg(test)]
fn fetch(url: &str) -> Result<String, Error> {
    use std::{
        collections::hash_map::DefaultHasher,
        fs,
        hash::{Hash, Hasher},
    };

    let mut hasher = DefaultHasher::new();
    url.hash(&mut hasher);
    let hash = format!("{:x}", hasher.finish());

    let expected_filename = format!("./data/cache/par-{hash}");

    match fs::read_to_string(&expected_filename) {
        Ok(data) => Ok(data),
        Err(_) => {
            info!("Not finding url {url} in the cache, downloading...");
            let data = ureq::get(url).call()?.into_string()?;
            fs::write(expected_filename, &data).expect("Unable to write file");
            Ok(data)
        }
    }
}

#[cfg(test)]
mod tests {

    use std::{fs, sync::Arc};

    use crate::{
        law::{
            responsible::{contains, starts_with_number},
            Classifier, LawBuilder,
        },
        risparser::paragraph::Parser,
    };
    use pretty_assertions::assert_eq;

    // TODO: Instead of specific test cases, use this general method once there is a nice way to
    // add classifiers + custom removals/edits of law texts
    //
    //#[test]
    //fn paragraph() {
    //    let laws = fs::read_dir("./data/expected/overview/")
    //        .expect("No folder with expected overview files");

    //    for law in laws {
    //        let path = law.unwrap();
    //        let law_path = format!("{}", path.path().display());
    //        let law_id = path.file_name().into_string().unwrap().to_string();

    //        //TODO: Remove this if once all law texts pass
    //        if ["10001905".into()].contains(&law_id) {
    //            let expected_path = format!("./data/expected/par/{law_id}");

    //            let pars = fs::read_to_string(law_path).expect("Could not read file {file_path}.");
    //            let pars = pars.trim().split('\n').collect::<Vec<&str>>();

    //            let mut builder = LawBuilder::new("law");
    //            builder.add_classifier(Classifier::new("Abschnitt", Arc::new(&contains)).root());

    //            let mut parser = Parser::new();
    //            parser.add_string_to_remove(r#"<absatz typ="abs" ct="text" halign="j"><i>(§§ 1, 2, 3, 4,6, § 8 Abs. 1 und 2 des Gesetzes vom 16. Februar 1883, RGBl. Nr. 20, betreffend das Verfahren zum Zwecke der Todeserklärung und der Beweisführung des Todes, in der Fassung des Gesetzes vom 31. März 1918, RGBl. Nr. 129:)</i></absatz>"#);
    //            parser.add_string_to_remove(r#"<absatz typ="abs" ct="text" halign="j"><i>(§§ 1 bis 11 des Gesetzes über die Verschollenheit, die Todeserklärung und die Feststellung der Todeszeit vom 4. Juli 1939, Deutsches RGBl. I S. 1186:)</i></absatz>"#);
    //            parser.add_string_to_remove(r#"<absatz typ="abs" ct="text" halign="j"><i>(§ 56 Abs. 3 des Gesetzes vom 4. Juli 1939, Deutsches RGBl. I S. 1186:)</i></absatz>"#);
    //            parser.add_string_to_remove(r#"<absatz typ="abs" ct="text" halign="j"><i>(§ 10 des Gesetzes vom 16. Februar 1883, RGBl. Nr. 20, betreffend das Verfahren zum Zwecke der Todeserklärung und der Beweisführung des Todes, in der Fassung des Gesetzes vom 31. März 1918, RGBl. Nr. 129:)</i></absatz>"#);
    //            parser.add_string_to_remove(r#"<absatz typ="abs" ct="text" halign="j"><i>(§§ 10a, 10b und 10c des Gesetzes vom 16. Februar 1883, RGBl. Nr. 20, betreffend das Verfahren zum Zwecke der Todeserklärung und der Beweisführung des Todes, in der Fassung des Gesetzes vom 31. März 1918, RGBl. Nr. 129:)</i></absatz>"#);
    //            parser.add_string_to_remove("<i>");
    //            parser.add_string_to_remove("</i>");
    //            for par in pars {
    //                let cont = parser.parse(par, &mut builder).unwrap();
    //                if !cont {
    //                    break;
    //                }
    //            }

    //            let actual = builder.history;

    //            let expected = fs::read_to_string(&expected_path)
    //                .expect(&format!("Could not read file {expected_path}."));
    //            let expected = expected.trim().split('\n').collect::<Vec<&str>>();

    //            assert_eq!(actual, expected);
    //        }
    //    }
    //}

    fn test(law_id: &str, builder: &mut LawBuilder, parser: Parser) {
        let paragraph_path = format!("./data/expected/overview/{law_id}");
        let expected_path = format!("./data/expected/par/{law_id}");

        let pars =
            fs::read_to_string(paragraph_path).expect("Could not read file {paragraph_path}.");
        let pars = pars.trim().split('\n').collect::<Vec<&str>>();

        for par in pars {
            println!("{par}");
            let cont = parser.parse(par, builder).unwrap();
            if !cont {
                break;
            }
        }

        let actual = &builder.history;

        println!("{actual:?}");

        let expected = fs::read_to_string(&expected_path)
            .expect(&format!("Could not read file {expected_path}."));
        let expected = expected.trim().split('\n').collect::<Vec<&str>>();

        assert_eq!(actual, &expected);
    }

    #[test]
    fn teg() {
        let law_id = "10001905";
        let mut builder = LawBuilder::new("law");
        builder.add_classifier(Classifier::new("Abschnitt", Arc::new(&contains)).root());

        let mut parser = Parser::new();
        parser.add_string_to_remove(r#"<absatz typ="abs" ct="text" halign="j"><i>(§§ 1, 2, 3, 4,6, § 8 Abs. 1 und 2 des Gesetzes vom 16. Februar 1883, RGBl. Nr. 20, betreffend das Verfahren zum Zwecke der Todeserklärung und der Beweisführung des Todes, in der Fassung des Gesetzes vom 31. März 1918, RGBl. Nr. 129:)</i></absatz>"#);
        parser.add_string_to_remove(r#"<absatz typ="abs" ct="text" halign="j"><i>(§§ 1 bis 11 des Gesetzes über die Verschollenheit, die Todeserklärung und die Feststellung der Todeszeit vom 4. Juli 1939, Deutsches RGBl. I S. 1186:)</i></absatz>"#);
        parser.add_string_to_remove(r#"<absatz typ="abs" ct="text" halign="j"><i>(§ 56 Abs. 3 des Gesetzes vom 4. Juli 1939, Deutsches RGBl. I S. 1186:)</i></absatz>"#);
        parser.add_string_to_remove(r#"<absatz typ="abs" ct="text" halign="j"><i>(§ 10 des Gesetzes vom 16. Februar 1883, RGBl. Nr. 20, betreffend das Verfahren zum Zwecke der Todeserklärung und der Beweisführung des Todes, in der Fassung des Gesetzes vom 31. März 1918, RGBl. Nr. 129:)</i></absatz>"#);
        parser.add_string_to_remove(r#"<absatz typ="abs" ct="text" halign="j"><i>(§§ 10a, 10b und 10c des Gesetzes vom 16. Februar 1883, RGBl. Nr. 20, betreffend das Verfahren zum Zwecke der Todeserklärung und der Beweisführung des Todes, in der Fassung des Gesetzes vom 31. März 1918, RGBl. Nr. 129:)</i></absatz>"#);
        parser.add_string_to_remove("<i>");
        parser.add_string_to_remove("</i>");

        test(law_id, &mut builder, parser);
    }

    #[test]
    fn mschg() {
        let law_id = "10002180";
        let mut builder = LawBuilder::new("law");
        builder.add_classifier(Classifier::new("Abschnitt", Arc::new(&contains)).root());
        builder.add_classifier(Classifier::new("Number", Arc::new(&starts_with_number)));

        let mut parser = Parser::new();
        parser.add_string_to_remove("<i>");
        parser.add_string_to_remove("</i>");

        parser.add_string_to_replace("\u{a0}", " ");
        parser.add_string_to_replace("<super>bis</super>", "bis");
        parser.add_string_to_replace("<super>ter</super>", "ter");
        parser.add_string_to_replace("<gdash />", "-");
        parser.add_string_to_replace(
            "(Anm.: § 69 aufgehoben durch Art. 1 Z 12, BGBl. I Nr. 124/2017)",
            "<gldsym>§ 69.</gldsym>(Anm.: § 69 aufgehoben durch Art. 1 Z 12, BGBl. I Nr. 124/2017)",
        );

        test(law_id, &mut builder, parser);
    }

    #[test]
    fn stgb() {
        let law_id = "10002296";
        let mut builder = LawBuilder::new("law");
        builder.add_classifier(Classifier::new("Teil", Arc::new(&contains)).root());
        builder.add_classifier(Classifier::new("Abschnitt", Arc::new(&contains)));

        let mut parser = Parser::new();
        parser.add_string_to_remove("<n>");
        parser.add_string_to_remove("</n>");
        parser.add_string_to_remove("<i>");
        parser.add_string_to_remove("</i>");
        parser.add_string_to_remove("<b>");
        parser.add_string_to_remove("</b>");
        parser.add_string_to_remove("<super>");
        parser.add_string_to_remove("</super>");
        parser.add_string_to_remove(r#"<abstand ct="text" halign="l" />"#);

        parser.add_string_to_replace("\u{a0}", " ");
        parser.add_string_to_replace(
            r#"<ueberschrift typ="g1min" ct="text" halign="c">Allgemeine Bestimmungen</ueberschrift>"#,
            r#"<ueberschrift typ="g2" ct="text" halign="c">Allgemeine Bestimmungen</ueberschrift>"#
        );
        parser.add_string_to_replace(
            r#"<ueberschrift typ="para" ct="text" halign="c">1. Verwertungsrechte.</ueberschrift>"#,
            r#"<ueberschrift typ="g1" ct="text" halign="c">1. Verwertungsrechte.</ueberschrift>"#,
        );
        parser.add_string_to_replace(
	        r#"<absatz typ="abs" ct="text" halign="j">(3) Die Verjährungsfrist beträgt</absatz><absatz typ="erltext" ct="text" halign="j">zwanzig Jahre,</absatz><liste><schluss typ="e1" ct="text">wenn die Handlung zwar nicht mit lebenslanger Freiheitsstrafe, aber mit mehr als zehnjähriger Freiheitsstrafe bedroht ist;</schluss></liste><absatz typ="erltext" ct="text" halign="j">zehn Jahre,</absatz><liste><schluss typ="e1" ct="text">wenn die Handlung mit mehr als fünfjähriger, aber höchstens zehnjähriger Freiheitsstrafe bedroht ist;</schluss></liste><absatz typ="erltext" ct="text" halign="j">fünf Jahre,</absatz><liste><schluss typ="e1" ct="text">wenn die Handlung mit mehr als einjähriger, aber höchstens fünfjähriger Freiheitsstrafe bedroht ist;</schluss></liste><absatz typ="erltext" ct="text" halign="j">drei Jahre,</absatz><liste><schluss typ="e1" ct="text">wenn die Handlung mit mehr als sechsmonatiger, aber höchstens einjähriger Freiheitsstrafe bedroht ist;</schluss></liste><absatz typ="erltext" ct="text" halign="j">ein Jahr,</absatz><liste><schluss typ="e1" ct="text">wenn die Handlung mit nicht mehr als sechsmonatiger Freiheitsstrafe oder nur mit Geldstrafe bedroht ist.</schluss></liste>"#,
	        r#"<absatz typ="abs" ct="text" halign="j">(3) Die Verjährungsfrist beträgt zwanzig Jahre, wenn die Handlung zwar nicht mit lebenslanger Freiheitsstrafe, aber mit mehr als zehnjähriger Freiheitsstrafe bedroht ist; zehn Jahre, wenn die Handlung mit mehr als fünfjähriger, aber höchstens zehnjähriger Freiheitsstrafe bedroht ist; fünf Jahre, wenn die Handlung mit mehr als einjähriger, aber höchstens fünfjähriger Freiheitsstrafe bedroht ist; drei Jahre, wenn die Handlung mit mehr als sechsmonatiger, aber höchstens einjähriger Freiheitsstrafe bedroht ist; ein Jahr, wenn die Handlung mit nicht mehr als sechsmonatiger Freiheitsstrafe oder nur mit Geldstrafe bedroht ist.</absatz>"#,
	    );

        parser.add_string_to_replace(
	        r#"<absatz typ="abs" ct="text" halign="j">(3) Die Frist beträgt</absatz><absatz typ="erltext" ct="text" halign="j">fünfzehn Jahre,</absatz><liste><schlussteil ebene="0.5" art="normal" ct="text">wenn auf Freiheitsstrafe von mehr als einem Jahr, aber nicht mehr als zehn Jahren erkannt worden ist;</schlussteil></liste><absatz typ="erltext" ct="text" halign="j">zehn Jahre,</absatz><liste><schlussteil ebene="0.5" art="normal" ct="text">wenn auf Freiheitsstrafe von mehr als drei Monaten, aber nicht mehr als einem Jahr oder auf eine Geldstrafe unter Festsetzung einer Ersatzfreiheitsstrafe von mehr als drei Monaten erkannt worden ist;</schlussteil></liste><absatz typ="erltext" ct="text" halign="j">fünf Jahre</absatz><liste><schlussteil ebene="0.5" art="normal" ct="text">in allen übrigen Fällen.</schlussteil></liste>"#,
	        r#"<absatz typ="abs" ct="text" halign="j">(3) Die Frist beträgt fünfzehn Jahre, wenn auf Freiheitsstrafe von mehr als einem Jahr, aber nicht mehr als zehn Jahren erkannt worden ist; zehn Jahre, wenn auf Freiheitsstrafe von mehr als drei Monaten, aber nicht mehr als einem Jahr oder auf eine Geldstrafe unter Festsetzung einer Ersatzfreiheitsstrafe von mehr als drei Monaten erkannt worden ist; fünf Jahre in allen übrigen Fällen.</absatz>"#,
	    );
        parser.add_string_to_replace("\u{a0}", " ");
        parser.add_string_to_replace(
            r#"<ueberschrift typ="g1min" ct="text" halign="c">Strafbare Handlungen gegen Leib und Leben</ueberschrift>"#,
	    r#"<ueberschrift typ="g2" ct="text" halign="c">Strafbare Handlungen gegen Leib und Leben</ueberschrift>"#,
        );
        parser.add_string_to_replace("<gdash />", "-");

        test(law_id, &mut builder, parser);
    }

    #[test]
    fn kschg() {
        let law_id = "10002462";
        let mut builder = LawBuilder::new("law");
        builder.add_classifier(Classifier::new("Hauptstück", Arc::new(&contains)).root());
        builder.add_classifier(Classifier::new("Abschnitt", Arc::new(&contains)));

        let mut parser = Parser::new();
        parser.add_string_to_remove("<i>");
        parser.add_string_to_remove("</i>");

        test(law_id, &mut builder, parser);
    }

    #[test]
    fn vvg() {
        let law_id = "20011654";
        let mut builder = LawBuilder::new("law");
        builder.add_classifier(Classifier::new("Abschnitt", Arc::new(&contains)).root());

        let parser = Parser::new();

        test(law_id, &mut builder, parser);
    }
    #[test]
    fn urhg() {
        let law_id = "10001848";
        let mut builder = LawBuilder::new("law");
        builder.add_classifier(Classifier::new("Hauptstück", Arc::new(&contains)).root());
        builder.add_classifier(Classifier::new("Abschnitt", Arc::new(&contains)));
        builder.add_classifier(Classifier::new("Number", Arc::new(&starts_with_number)));

        let mut parser = Parser::new();
        parser.add_string_to_replace("<i>.</i>", ".");
        parser.add_string_to_replace("<gdash />", "-");
        parser.add_string_to_replace(
            r#"<liste><schlussteil ebene="0" art="normal" ct="text">(2) Einer Rundfunksendung steht es gleich, wenn ein Werk von einer im In- oder im Ausland gelegenen Stelle aus der Öffentlichkeit im Inland, ähnlich wie durch Rundfunk, aber mit Hilfe von Leitungen wahrnehmbar gemacht wird.</schlussteil></liste>"#,
            r#"<absatz typ="abs" ct="text" halign="j">(2) Einer Rundfunksendung steht es gleich, wenn ein Werk von einer im In- oder im Ausland gelegenen Stelle aus der Öffentlichkeit im Inland, ähnlich wie durch Rundfunk, aber mit Hilfe von Leitungen wahrnehmbar gemacht wird.</absatz>"#,
        );
        parser.add_string_to_replace("\u{a0}", " ");

        test(law_id, &mut builder, parser);
    }

    //  TODO: 10001622(ABGB)
    //  https://www.ris.bka.gv.at/GeltendeFassung.wxe?Abfrage=Bundesnormen&Gesetzesnummer=10002180
}
